"""
Setup:

1, Go to the Admin Dashboard in Okta
2. Go to Security > API > Tokens
3. Click on "Create Token"

"""

from typing import Annotated
from httpx import Client, Response
from pydantic import BaseModel

from admyral.action import action, ArgumentMetadata
from admyral.context import ctx
from admyral.typings import JsonValue
from admyral.secret.secret import register_secret


@register_secret(secret_type="Okta")
class OktaSecret(BaseModel):
    domain: str
    api_key: str


# TODO: OAuth2: https://developer.okta.com/docs/guides/implement-oauth-for-okta-serviceapp/main/
def get_okta_client(secret: OktaSecret) -> Client:
    return Client(
        base_url=f"https://{secret.domain}/api/v1",
        headers={
            "Authorization": f"SSWS {secret.api_key}",
            "Accept": "application/json",
        },
    )


def _get_next_link(response: Response) -> str | None:
    next_link = [
        header.split(";")[0][1:-1]
        for header in response.headers.get_list("link")
        if header.endswith('rel="next"')
    ]
    return next_link[0] if len(next_link) > 0 else None


@action(
    display_name="List Events",
    display_namespace="Okta",
    description="List events from Okta",
    secrets_placeholders=["OKTA_SECRET"],
)
def list_okta_events(
    user_id: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="User ID",
            description="The user ID to list events for.",
        ),
    ] = None,
    start_time: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Start Time",
            description="The start time for the cases to list. Must be in ISO 8601 format (YYYY-MM-DDTHH:MM:SSZ).",
        ),
    ] = "1970-01-01T00:00:00Z",
    end_time: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="End Time",
            description="The end time for the cases to list. Must be in ISO 8601 format (YYYY-MM-DDTHH:MM:SSZ).",
        ),
    ] = "2100-01-01T00:00:00Z",
    limit: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Limit",
            description="The maximum number of cases to list.",
        ),
    ] = 1000,
) -> list[dict[str, JsonValue]]:
    # Note: Bounded Request - Polling Request use case not supported
    # https://developer.okta.com/docs/reference/api/system-log/#list-events

    secret = ctx.get().secrets.get("OKTA_SECRET")
    secret = OktaSecret.model_validate(secret)

    with get_okta_client(secret) as client:
        params = {
            "limit": min(limit, 1000),
            "since": start_time,
            "until": end_time,
        }
        if user_id:
            params["filter"] = f'actor.id eq "{user_id}"'

        response = client.get("/logs", params=params)
        response.raise_for_status()

        events = response.json()

        # handle pagination
        prev_num_events = len(events)
        base_url = str(client.base_url)
        while link := _get_next_link(response):
            response = client.get(link.replace(base_url, ""))
            response.raise_for_status()
            events.extend(response.json())
            if len(events) >= limit or prev_num_events == len(events):
                break
            prev_num_events = len(events)

        return events[:limit]


@action(
    display_name="Search Users",
    display_namespace="Okta",
    description="Search Users from Okta",
    secrets_placeholders=["OKTA_SECRET"],
)
def okta_search_users(
    search: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Search",
            description="The search query to filter users.",
        ),
    ] = None,
    limit: Annotated[
        int | None,
        ArgumentMetadata(
            display_name="Limit",
            description="The maximum number of users to list.",
        ),
    ] = None,
) -> list[dict[str, JsonValue]]:
    # https://developer.okta.com/docs/api/openapi/okta-management/management/tag/User/#tag/User/operation/listUsers
    secret = ctx.get().secrets.get("OKTA_SECRET")
    secret = OktaSecret.model_validate(secret)

    with get_okta_client(secret) as client:
        params = {
            "search": search,
            "limit": min(limit or 200, 200),  # Okta's maximum limit per request is 200
        }

        response = client.get("/users", params=params)
        response.raise_for_status()

        users = response.json()

        # handle pagination
        prev_num_users = len(users)
        base_url = str(client.base_url)
        while link := _get_next_link(response):
            response = client.get(link.replace(base_url, ""))
            response.raise_for_status()
            users.extend(response.json())
            if limit is not None and (
                len(users) >= limit or prev_num_users == len(users)
            ):
                break
            prev_num_users = len(users)

        return users[:limit] if limit is not None else users


@action(
    display_name="Get All User Types",
    display_namespace="Okta",
    description="Retrieve all user types from Okta",
    secrets_placeholders=["OKTA_SECRET"],
)
def okta_get_all_user_types() -> list[dict[str, JsonValue]]:
    # https://developer.okta.com/docs/api/openapi/okta-management/management/tag/UserType/#tag/UserType/operation/listUserTypes
    secret = ctx.get().secrets.get("OKTA_SECRET")
    secret = OktaSecret.model_validate(secret)

    with get_okta_client(secret) as client:
        response = client.get("/meta/types/user")
        response.raise_for_status()

        user_types = response.json()

        return user_types


@action(
    display_name="Get Okta Logs",
    display_namespace="Okta",
    description="Retrieve Okta system logs",
    secrets_placeholders=["OKTA_SECRET"],
)
def get_okta_logs(
    query: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Query",
            description="The query to filter logs. For example, 'logout'",
        ),
    ] = None,
    start_time: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Start Time",
            description="The start time for the logs to retrieve. Must be in ISO 8601 format (YYYY-MM-DDTHH:MM:SSZ).",
        ),
    ] = None,
    end_time: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="End Time",
            description="The end time for the logs to retrieve. Must be in ISO 8601 format (YYYY-MM-DDTHH:MM:SSZ).",
        ),
    ] = None,
    limit: Annotated[
        int,
        ArgumentMetadata(
            display_name="Limit",
            description="The maximum number of logs to retrieve.",
        ),
    ] = 1000,
) -> list[dict[str, JsonValue]]:
    # https://developer.okta.com/docs/api/openapi/okta-management/management/tag/SystemLog/#tag/SystemLog/operation/listLogEvents
    secret = ctx.get().secrets.get("OKTA_SECRET")
    secret = OktaSecret.model_validate(secret)

    with get_okta_client(secret) as client:
        params = {
            "limit": min(limit, 1000),
        }
        if query:
            params["q"] = query
        if start_time:
            params["since"] = start_time
        if end_time:
            params["until"] = end_time

        response = client.get("/logs", params=params)
        response.raise_for_status()

        logs = response.json()

        # handle pagination
        prev_num_logs = len(logs)
        base_url = str(client.base_url)
        while link := _get_next_link(response):
            response = client.get(link.replace(base_url, ""))
            response.raise_for_status()
            logs.extend(response.json())
            if len(logs) >= limit or prev_num_logs == len(logs):
                break
            prev_num_logs = len(logs)

        return logs[:limit]



from typing import Annotated

from admyral.action import action, ArgumentMetadata
from admyral.context import ctx
from admyral.typings import JsonValue
from admyral.actions.integrations.shared.steampipe import run_steampipe_query
from admyral.actions.integrations.shared.aws import AWSSecret


def get_aws_secret() -> AWSSecret:
    secret = ctx.get().secrets.get("AWS_SECRET")
    return AWSSecret.model_validate(secret)


@action(
    display_name="Steampipe Query",
    display_namespace="AWS",
    description="Query AWS using Steampipe",
    secrets_placeholders=["AWS_SECRET"],
)
def steampipe_query_aws(
    query: Annotated[
        str,
        ArgumentMetadata(
            display_name="Query",
            description="The Steampipe query to execute",
        ),
    ],
) -> JsonValue:
    secret = get_aws_secret()
    return run_steampipe_query(
        query, secret.aws_access_key_id, secret.aws_secret_access_key
    )


@action(
    display_name="Control: S3 Bucket Access Logging",
    display_namespace="AWS",
    description="S3 server access logging provides a method to monitor the network for potential cybersecurity events.",
    secrets_placeholders=["AWS_SECRET"],
)
def aws_s3_bucket_logging_enabled(
    query: Annotated[
        str,
        ArgumentMetadata(
            display_name="Query",
            description="The query representing the control",
        ),
    ] = """select
    arn as resource,
    case
    when logging->'TargetBucket' is null then 'alarm'
    else 'ok'
    end as status,
    case
    when logging->'TargetBucket' is null then title || ' logging disabled.'
    else title || ' logging enabled.'
    end as reason
from
    aws_s3_bucket;
    """,
) -> JsonValue:
    secret = get_aws_secret()
    return run_steampipe_query(
        query, secret.aws_access_key_id, secret.aws_secret_access_key
    )


@action(
    display_name="Control: S3 Bucket Enforce SSL",
    display_namespace="AWS",
    description="To help protect data in transit, ensure that your S3 buckets require requests to use SSL.",
    secrets_placeholders=["AWS_SECRET"],
)
def aws_s3_bucket_enforce_ssl(
    query: Annotated[
        str,
        ArgumentMetadata(
            display_name="Query",
            description="The query representing the control",
        ),
    ] = """with ssl_ok as (
    select
        distinct name,
        arn,
        'ok' as status
    from
        aws_s3_bucket,
        jsonb_array_elements(policy_std -> 'Statement') as s,
        jsonb_array_elements_text(s -> 'Principal' -> 'AWS') as p,
        jsonb_array_elements_text(s -> 'Action') as a,
        jsonb_array_elements_text(s -> 'Resource') as r,
        jsonb_array_elements_text(
            s -> 'Condition' -> 'Bool' -> 'aws:securetransport'
        ) as ssl
    where
        p = '*'
        and s ->> 'Effect' = 'Deny'
        and ssl :: bool = false
)
select
    b.arn as resource,
    case
        when ok.status = 'ok' then 'ok'
        else 'alarm'
    end status,
    case
        when ok.status = 'ok' then b.name || ' bucket policy enforces HTTPS.'
        else b.name || ' bucket policy does not enforce HTTPS.'
    end reason
from
    aws_s3_bucket as b
left join ssl_ok as ok on ok.name = b.name;
    """,
) -> JsonValue:
    secret = get_aws_secret()
    return run_steampipe_query(
        query, secret.aws_access_key_id, secret.aws_secret_access_key
    )


@action(
    display_name="Control: S3 Bucket Restrict Public Read Access",
    display_namespace="AWS",
    description="S3 public read access should be blocked.",
    secrets_placeholders=["AWS_SECRET"],
)
def aws_s3_restrict_public_read_access(
    query: Annotated[
        str,
        ArgumentMetadata(
            display_name="Query",
            description="The query representing the control",
        ),
    ] = """with public_acl as (
    select
        distinct name
    from
        aws_s3_bucket,
        jsonb_array_elements(acl->'Grants') as grants
    where
        (grants->'Grantee'->>'URI' = 'http://acs.amazonaws.com/groups/global/AllUsers'
        or grants->'Grantee'->>'URI' = 'http://acs.amazonaws.com/groups/global/AuthenticatedUsers')
        and (
          grants->>'Permission' = 'FULL_CONTROL'
          or grants->>'Permission' = 'READ_ACP'
          or grants->>'Permission' = 'READ'
        )
),read_access_policy as (
    select
        distinct name
    from
        aws_s3_bucket,
        jsonb_array_elements(policy_std->'Statement') as s,
        jsonb_array_elements_text(s->'Action') as action
    where
        s->>'Effect' = 'Allow'
        and (
            s->'Principal'->'AWS' = '["*"]'
            or s->>'Principal' = '*'
        )
        and (
            action = '*'
            or action = '*:*'
            or action = 's3:*'
            or action ilike 's3:get%'
            or action ilike 's3:list%'
        )
)
select
    b.arn as resource,
    case
        when (block_public_acls or a.name is null) and not bucket_policy_is_public then 'ok'
        when (block_public_acls or a.name is null) and (bucket_policy_is_public and block_public_policy) then 'ok'
        when (block_public_acls or a.name is null) and (bucket_policy_is_public and p.name is null) then 'ok'
        else 'alarm'
    end as status,
    case
        when (block_public_acls or a.name is null) and not bucket_policy_is_public then b.title || ' not publicly readable.'
        when (block_public_acls or a.name is null) and (bucket_policy_is_public and block_public_policy) then b.title || ' not publicly readable.'
        when (block_public_acls or a.name is null) and (bucket_policy_is_public and p.name is null) then  b.title || ' not publicly readable.'
        else b.title || ' publicly readable.'
    end as reason
from
    aws_s3_bucket as b
    left join public_acl as a on b.name = a.name
    left join read_access_policy as p on b.name = p.name;
    """,
) -> JsonValue:
    secret = get_aws_secret()
    return run_steampipe_query(
        query, secret.aws_access_key_id, secret.aws_secret_access_key
    )


@action(
    display_name="Control: S3 Bucket Restrict Public Write Access",
    display_namespace="AWS",
    description="S3 public write access should be blocked.",
    secrets_placeholders=["AWS_SECRET"],
)
def aws_s3_restrict_public_write_access(
    query: Annotated[
        str,
        ArgumentMetadata(
            display_name="Query",
            description="The query representing the control",
        ),
    ] = """with public_acl as (
    select
        distinct name
    from
        aws_s3_bucket,
        jsonb_array_elements(acl -> 'Grants') as grants
    where
        (grants -> 'Grantee' ->> 'URI' = 'http://acs.amazonaws.com/groups/global/AllUsers'
        or grants -> 'Grantee' ->> 'URI' = 'http://acs.amazonaws.com/groups/global/AuthenticatedUsers')
        and (
          grants ->> 'Permission' = 'FULL_CONTROL'
          or grants ->> 'Permission' = 'WRITE_ACP'
          or grants ->> 'Permission' = 'WRITE'
        )
), write_access_policy as (
    select
        distinct name
    from
        aws_s3_bucket,
        jsonb_array_elements(policy_std -> 'Statement') as s,
        jsonb_array_elements_text(s -> 'Action') as action
    where
        s ->> 'Effect' = 'Allow'
        and (
        s -> 'Principal' -> 'AWS' = '["*"]'
        or s ->> 'Principal' = '*'
        )
        and (
            action = '*'
            or action = '*:*'
            or action = 's3:*'
            or action ilike 's3:put%'
            or action ilike 's3:delete%'
            or action ilike 's3:create%'
            or action ilike 's3:update%'
            or action ilike 's3:replicate%'
            or action ilike 's3:restore%'
        )
)
select
    b.arn as resource,
    case
        when (block_public_acls or a.name is null) and not bucket_policy_is_public then 'ok'
        when (block_public_acls or a.name is null) and (bucket_policy_is_public and block_public_policy) then 'ok'
        when bucket_policy_is_public and p.name is null then 'ok'
        else 'alarm'
    end status,
    case
        when (block_public_acls or a.name is null ) and not bucket_policy_is_public then b.title || ' not publicly writable.'
        when (block_public_acls or a.name is null) and (bucket_policy_is_public and block_public_policy) then b.title || ' not publicly writable.'
        when (block_public_acls or a.name is null) and (bucket_policy_is_public and p.name is null) then b.title || ' not publicly writable.'
        else b.title || ' publicly writable.'
    end reason
from
    aws_s3_bucket as b
    left join public_acl as a on b.name = a.name
    left join write_access_policy as p on b.name = p.name;
    """,
) -> JsonValue:
    secret = get_aws_secret()
    return run_steampipe_query(
        query, secret.aws_access_key_id, secret.aws_secret_access_key
    )


@action(
    display_name="Control: S3 Bucket Default Encryption Enabled",
    display_namespace="AWS",
    description="S3 bucket default encryption should be enabled.",
    secrets_placeholders=["AWS_SECRET"],
)
def aws_s3_default_encryption_enabled(
    query: Annotated[
        str,
        ArgumentMetadata(
            display_name="Query",
            description="The query representing the control",
        ),
    ] = """select
    arn as resource,
    case
        when server_side_encryption_configuration is not null then 'ok'
        else 'alarm'
    end status,
    case
        when server_side_encryption_configuration is not null then name || ' default encryption enabled.'
        else name || ' default encryption disabled.'
    end reason
from
    aws_s3_bucket;
    """,
) -> JsonValue:
    secret = get_aws_secret()
    return run_steampipe_query(
        query, secret.aws_access_key_id, secret.aws_secret_access_key
    )


@action(
    display_name="List IAM Users",
    display_namespace="AWS",
    description="List AWS IAM users.",
    secrets_placeholders=["AWS_SECRET"],
)
def aws_list_iam_users(
    query: Annotated[
        str,
        ArgumentMetadata(
            display_name="Query",
            description="The query for fetching IAM users.",
        ),
    ] = """select
  account_id,
  akas,
  arn,
  attached_policy_arns,
  create_date,
  groups,
  inline_policies,
  inline_policies_std,
  login_profile,
  mfa_devices,
  mfa_enabled,
  name,
  partition,
  password_last_used,
  path,
  permissions_boundary_arn,
  permissions_boundary_type,
  region,
  user_id
from
  aws_iam_user;
    """,
) -> JsonValue:
    secret = get_aws_secret()
    return run_steampipe_query(
        query, secret.aws_access_key_id, secret.aws_secret_access_key
    )



import click
import os
import json

from admyral.cli.cli import cli
from admyral.compiler.workflow_compiler import WorkflowCompiler
from admyral.models import TriggerStatus
from admyral.client import AdmyralClient
from admyral.utils.telemetry import capture


@cli.group()
def workflow() -> None:
    """Workflow Management"""


# TODO: add option for pushing all used Python actions automatically
@workflow.command(
    "push",
    help="Push a workflow to Admyral",
)
@click.argument("workflow_name", type=str)
@click.option(
    "--file",
    "-f",
    type=str,
    help="Path to the Python file containing the workflow",
)
@click.option(
    "--activate",
    is_flag=True,
    help="Activate the workflow after pushing it to Admyral",
)
@click.pass_context
def push(ctx: click.Context, workflow_name: str, file: str, activate: bool) -> None:
    """Push workflow to Admyral"""
    capture(event_name="workflow:push")
    client: AdmyralClient = ctx.obj

    # compile the workflow
    if not os.path.exists(file):
        click.echo(f"File {file} not found.")
        return
    with open(file, "r") as f:
        workflow_code = f.read()
    workflow_dag = WorkflowCompiler().compile_from_module(workflow_code, workflow_name)

    # Push workflow to Admyral
    try:
        workflow_push_response = client.push_workflow(
            workflow_name=workflow_name, workflow_dag=workflow_dag, is_active=activate
        )
    except Exception as e:
        click.echo(f"Failed to push workflow {workflow_name}.")
        click.echo(f"Error: {e}")
        return

    click.echo(f"Workflow {workflow_name} pushed successfully.")

    if workflow_push_response.webhook_id:
        click.echo(f"Webhook ID: {workflow_push_response.webhook_id}")
        click.echo(f"Webhook Secret: {workflow_push_response.webhook_secret}")


@workflow.command(
    "trigger",
    help='Trigger a workflow. Example: admyral workflow trigger my_workflow --payload \'{"ip": "127.0.0.1"}\'',
)
@click.argument("workflow_name", type=str)
@click.option(
    "--payload",
    "-p",
    type=str,
    help='Payload to send to the workflow. Example: \'{"ip": "127.0.0.1"}\'',
)
@click.pass_context
def trigger(ctx: click.Context, workflow_name: str, payload: str | None) -> None:
    """Trigger workflow execution"""
    capture(event_name="workflow:trigger")
    client: AdmyralClient = ctx.obj
    payload = json.loads(payload) if payload else None
    try:
        response = client.trigger_workflow(workflow_name, payload)
        match response.status:
            case TriggerStatus.SUCCESS:
                click.echo(f"Workflow {workflow_name} triggered successfully.")
            case TriggerStatus.INACTIVE:
                click.echo(f"Workflow {workflow_name} is deactivated.")

    except Exception as e:
        if "is not active" in str(e):
            click.echo(f"Workflow {workflow_name} is not active.")
        else:
            click.echo(f"Failed to trigger workflow {workflow_name}.")
            click.echo(f"Error: {e}")


@workflow.command(
    "activate",
    help="Activate a workflow",
)
@click.argument("workflow_name", type=str)
@click.pass_context
def activate_workflow(ctx: click.Context, workflow_name: str) -> None:
    """Activate a workflow"""
    capture(event_name="workflow:activate")
    client: AdmyralClient = ctx.obj
    try:
        client.activate_workflow(workflow_name)
    except Exception as e:
        click.echo(f"Failed to activate workflow {workflow_name}.")
        click.echo(f"Error: {e}")
        return
    click.echo(f"Activated workflow {workflow_name} successfully.")


@workflow.command(
    "deactivate",
    help="Deactivate a workflow",
)
@click.argument("workflow_name", type=str)
@click.pass_context
def deactivate_workflow(ctx: click.Context, workflow_name: str) -> None:
    """Deactivate a workflow"""
    capture(event_name="workflow:deactivate")
    client: AdmyralClient = ctx.obj
    try:
        client.deactivate_workflow(workflow_name)
    except Exception as e:
        click.echo(f"Failed to deactivate workflow {workflow_name}.")
        click.echo(f"Error: {e}")
        return
    click.echo(f"Deactivated workflow {workflow_name} successfully.")



from typing import Annotated
from httpx import Client
from pydantic import BaseModel

from admyral.action import action, ArgumentMetadata
from admyral.context import ctx
from admyral.typings import JsonValue
from admyral.secret.secret import register_secret


@register_secret(secret_type="Kandji")
class KandjiSecret(BaseModel):
    api_url: str
    api_token: str


def get_kandji_client(secret: KandjiSecret) -> Client:
    return Client(
        base_url=f"https://{secret.api_url}/api/v1",
        headers={
            "Authorization": f"Bearer {secret.api_token}",
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
    )


def _kandji_get_api_with_pagination(url: str, data_access_key: str | None = None):
    secret = ctx.get().secrets.get("KANDJI_SECRET")
    secret = KandjiSecret.model_validate(secret)

    with get_kandji_client(secret) as client:
        offset = 0
        max_limit_per_page = 300

        out = []

        while True:
            response = client.get(
                url=url,
                params={"limit": max_limit_per_page, "offset": offset},
            )
            response.raise_for_status()

            result = response.json()
            if data_access_key is None:
                new_data = result
            else:
                new_data = result[data_access_key]
            out.extend(new_data)

            offset += len(new_data)
            if len(new_data) < max_limit_per_page:
                break

        return out


def _kandji_get_api(url: str) -> dict[str, JsonValue]:
    secret = ctx.get().secrets.get("KANDJI_SECRET")
    secret = KandjiSecret.model_validate(secret)

    with get_kandji_client(secret) as client:
        response = client.get(url=url)
        response.raise_for_status()
        return response.json()


@action(
    display_name="List Devices",
    display_namespace="Kandji",
    description="List devices managed by Kandji",
    secrets_placeholders=["KANDJI_SECRET"],
)
def list_kandji_devices() -> list[dict[str, JsonValue]]:
    # https://api-docs.kandji.io/#78209960-31a7-4e3b-a2c0-95c7e65bb5f9
    return _kandji_get_api_with_pagination(url="/devices")


@action(
    display_name="Get Device Details",
    display_namespace="Kandji",
    description="List devices managed by Kandji",
    secrets_placeholders=["KANDJI_SECRET"],
)
def get_kandji_device_details(
    device_id: Annotated[
        str,
        ArgumentMetadata(display_name="Device ID", description="The ID of the device"),
    ],
) -> list[dict[str, JsonValue]]:
    # https://api-docs.kandji.io/#efa2170d-e5f7-4b97-8f4c-da6f84ba58b5
    return _kandji_get_api(url=f"/devices/{device_id}/details")


@action(
    display_name="List Unencrypted Devices",
    display_namespace="Kandji",
    description="List devices managed by Kandji which don't have disk encryption enabled on all volumes",
    secrets_placeholders=["KANDJI_SECRET"],
)
def list_kandji_unencrypted_devices() -> list[dict[str, JsonValue]]:
    # https://api-docs.kandji.io/#78209960-31a7-4e3b-a2c0-95c7e65bb5f9

    devices = _kandji_get_api_with_pagination(url="/devices")

    unencrypted_devices = []
    for device in devices:
        device_details = _kandji_get_api(url=f"/devices/{device["device_id"]}/details")
        if any(volume["encrypted"] == "No" for volume in device_details["volumes"]):
            unencrypted_devices.append(device)

    return unencrypted_devices


@action(
    display_name="Get Device Apps",
    display_namespace="Kandji",
    description="List the installed apps of a Kandji managed device",
    secrets_placeholders=["KANDJI_SECRET"],
)
def get_kandji_device_apps(
    device_id: Annotated[
        str,
        ArgumentMetadata(display_name="Device ID", description="The ID of the device"),
    ],
) -> list[dict[str, JsonValue]]:
    # https://api-docs.kandji.io/#f8cd9733-89b6-40f0-a7ca-76829c6974df
    return _kandji_get_api_with_pagination(
        url=f"/devices/{device_id}/apps", data_access_key="apps"
    )


@action(
    display_name="Application Firewall",
    display_namespace="Kandji",
    description="Get Application Firewall details for macOS.",
    secrets_placeholders=["KANDJI_SECRET"],
)
def get_kandji_application_firewall() -> list[dict[str, JsonValue]]:
    # https://api-docs.kandji.io/#0d3abc6c-a5a5-4fe6-b7d4-19b1287eaf91

    return _kandji_get_api_with_pagination(
        url="/prism/application_firewall", data_access_key="data"
    )


@action(
    display_name="Desktop and Screensaver",
    display_namespace="Kandji",
    description="Get Desktop and Screensaver details for macOS.",
    secrets_placeholders=["KANDJI_SECRET"],
)
def get_kandji_desktop_and_screensaver() -> list[dict[str, JsonValue]]:
    # https://api-docs.kandji.io/#7f6f6813-6522-4249-9f6a-3d0c479e2bbe
    return _kandji_get_api_with_pagination(
        url="/prism/desktop_and_screensaver", data_access_key="data"
    )


@action(
    display_name="Get Library Item Statuses",
    display_namespace="Kandji",
    description="This endpoint retrieves the statuses related to a specific library item.",
    secrets_placeholders=["KANDJI_SECRET"],
)
def get_kandji_library_item_statuses(
    library_item_id: Annotated[
        str,
        ArgumentMetadata(
            display_name="Library Item ID", description="The ID of the library item"
        ),
    ],
) -> list[dict[str, JsonValue]]:
    # https://api-docs.kandji.io/#478764c4-638c-416c-b44c-3685a2f7b441
    return _kandji_get_api_with_pagination(
        url=f"/library/library-items/{library_item_id}/status",
        data_access_key="results",
    )



from sqlmodel import Field, Relationship
from sqlalchemy import TEXT, JSON, ForeignKeyConstraint, TIMESTAMP
from typing import TYPE_CHECKING
from datetime import datetime

from admyral.db.schemas.base_schemas import BaseSchema
from admyral.models import WorkflowRun, WorkflowRunStep, WorkflowRunMetadata
from admyral.typings import JsonValue

if TYPE_CHECKING:
    from admyral.db.schemas.workflow_schemas import WorkflowSchema


class WorkflowRunSchema(BaseSchema, table=True):
    """
    Schema for Workflow Run
    """

    __tablename__ = "workflow_runs"

    __table_args__ = (
        ForeignKeyConstraint(
            ["workflow_id"],
            ["workflows.workflow_id"],
            ondelete="CASCADE",
        ),
    )

    # primary keys
    run_id: str = Field(sa_type=TEXT(), primary_key=True)

    # foreign keys
    workflow_id: str = Field(sa_type=TEXT())

    # other fields
    source_name: str = Field(sa_type=TEXT())
    completed_at: datetime | None = Field(sa_type=TIMESTAMP(), nullable=True)
    failed_at: datetime | None = Field(sa_type=TIMESTAMP(), nullable=True)
    canceled_at: datetime | None = Field(sa_type=TIMESTAMP(), nullable=True)

    # relationship parents
    workflow: "WorkflowSchema" = Relationship(back_populates="workflow_runs")

    # relationship children
    steps: list["WorkflowRunStepsSchema"] = Relationship(
        back_populates="workflow_run",
        sa_relationship_kwargs=dict(cascade="all, delete"),
    )

    def to_model(self, include_resources: bool = False) -> WorkflowRun:
        workflow_run = WorkflowRun.model_validate(
            {
                "run_id": self.run_id,
                "trigger_id": self.trigger_id,
                "payload": self.payload,
                "created_at": self.created_at,
                "completed_at": self.completed_at,
                "failed_at": self.failed_at,
                "canceled_at": self.canceled_at,
            }
        )

        # if include_resources:
        #     # TODO: do we have access to self.action_executions???
        #     workflow_run.action_executions = [
        #         action_execution.to_model()
        #         for action_execution in self.action_executions
        #     ]

        return workflow_run

    def to_metadata(self) -> WorkflowRunMetadata:
        return WorkflowRunMetadata.model_validate(
            {
                "run_id": self.run_id,
                "created_at": self.created_at,
                "completed_at": self.completed_at,
                "failed_at": self.failed_at,
                "canceled_at": self.canceled_at,
            }
        )


class WorkflowRunStepsSchema(BaseSchema, table=True):
    """
    Schema for Workflow Run Step
    """

    __tablename__ = "workflow_run_steps"

    __table_args__ = (
        ForeignKeyConstraint(
            ["run_id"],
            ["workflow_runs.run_id"],
            ondelete="CASCADE",
        ),
    )

    # primary keys
    step_id: str = Field(sa_type=TEXT(), primary_key=True)

    # foreign keys
    run_id: str = Field(sa_type=TEXT())

    # other fields
    prev_step_id: str | None = Field(sa_type=TEXT(), nullable=True)
    action_type: str = Field(sa_type=TEXT())
    input_args: dict[str, JsonValue] | None = Field(sa_type=JSON(), nullable=True)
    logs: str | None = Field(sa_type=TEXT(), nullable=True)
    result: JsonValue = Field(sa_type=JSON(), nullable=True)
    error: str | None = Field(sa_type=TEXT(), nullable=True)

    # relationship parents
    workflow_run: WorkflowRunSchema = Relationship(back_populates="steps")

    def to_model(self, include_resources: bool = False) -> WorkflowRunStep:
        return WorkflowRunStep.model_validate(
            {
                "step_id": self.step_id,
                "action_type": self.action_type,
                "prev_step_id": self.prev_step_id,
                "logs": self.logs,
                "result": self.result,
                "error": self.error,
                "input_args": self.input_args,
            }
        )



from sqlmodel import Field, ForeignKeyConstraint, Relationship
from sqlalchemy import TEXT
from typing import Optional, TYPE_CHECKING
import json

from admyral.db.schemas.base_schemas import BaseSchema
from admyral.models import EncryptedSecret

if TYPE_CHECKING:
    from admyral.db.schemas.auth_schemas import UserSchema


class SecretsSchema(BaseSchema, table=True):
    """
    Schema for Secrets
    """

    __tablename__ = "secrets"
    __table_args__ = (
        ForeignKeyConstraint(
            ["user_id"],
            ["User.id"],
            ondelete="CASCADE",
        ),
    )

    # primary keys
    secret_id: str = Field(sa_type=TEXT(), primary_key=True)

    # foreign keys
    user_id: str = Field(sa_type=TEXT(), primary_key=True, index=True)

    # other fields
    encrypted_secret: Optional[str] = Field(sa_type=TEXT(), nullable=True)
    """ Note: encrypted_secret is nullable because we may not have the secret stored in the database. """
    schema_json_serialized: str = Field(sa_type=TEXT())
    """ Serialized JSON Array of key names of the secret. """
    secret_type: Optional[str] = Field(sa_type=TEXT(), nullable=True)
    """ Type of the secret. """

    # relationship parent
    user: "UserSchema" = Relationship(back_populates="secrets")

    def to_model(self, include_resources: bool = False) -> EncryptedSecret:
        return EncryptedSecret.model_validate(
            {
                "secret_id": self.secret_id,
                "encrypted_secret": self.encrypted_secret,
                "secret_schema": json.loads(self.schema_json_serialized),
                "secret_type": self.secret_type,
            }
        )



from typing import Annotated, Literal
from httpx import Client
from pydantic import BaseModel

from admyral.action import action, ArgumentMetadata
from admyral.context import ctx
from admyral.typings import JsonValue
from admyral.secret.secret import register_secret


@register_secret(secret_type="AbuseIPDB")
class AbuseIPDBSecret(BaseModel):
    api_key: str


def get_abuseipdb_client(secret: AbuseIPDBSecret) -> Client:
    return Client(
        base_url="https://api.abuseipdb.com/api/v2",
        headers={
            "Key": secret.api_key,
            "Accept": "application/json",
        },
    )


@action(
    display_name="Check IP Address",
    display_namespace="AbuseIPDB",
    description="Check an IP address using AbuseIPDB",
    secrets_placeholders=["ABUSEIPDB_SECRET"],
)
def abuseipdb_analyze_ip(
    ip_address: Annotated[
        str,
        ArgumentMetadata(
            display_name="IP Address",
            description="The IP address (v4 or v6) to check",
        ),
    ],
    verbose: Annotated[
        Literal["yes", "no"] | None,
        ArgumentMetadata(
            display_name="Verbose",
            description="Whether to return verbose output",
        ),
    ] = "no",
    max_age_in_days: Annotated[
        int | None,
        ArgumentMetadata(
            display_name="Max Age in Days",
            description="The maximum age of the report in days",
        ),
    ] = 30,
) -> JsonValue:
    # https://docs.abuseipdb.com/#check-endpoint

    secret = ctx.get().secrets.get("ABUSEIPDB_SECRET")
    secret = AbuseIPDBSecret.model_validate(secret)

    with get_abuseipdb_client(secret) as client:
        params = {
            "ipAddress": ip_address,
        }
        if verbose == "yes":
            params["verbose"] = verbose
        if max_age_in_days and 1 <= max_age_in_days <= 365:
            params["maxAgeInDays"] = max_age_in_days
        response = client.get("/check", params=params)
        response.raise_for_status()
        return response.json().get("data", {})



from sqlmodel import Field, Relationship, ForeignKeyConstraint
from sqlalchemy import TEXT
from typing import TYPE_CHECKING

from admyral.db.schemas.base_schemas import BaseSchema

if TYPE_CHECKING:
    from admyral.db.schemas.auth_schemas import UserSchema


class ApiKeySchema(BaseSchema, table=True):
    """
    Schema for API Keys
    """

    __tablename__ = "api_keys"
    __table_args__ = (
        ForeignKeyConstraint(
            ["user_id"],
            ["User.id"],
            ondelete="CASCADE",
        ),
    )

    # primary keys
    id: str = Field(sa_type=TEXT(), primary_key=True)

    # foreign keys
    user_id: str = Field(sa_type=TEXT(), index=True)

    # other fields
    name: str = Field(sa_type=TEXT())
    key: str = Field(sa_type=TEXT(), index=True)

    # relationship parent
    user: "UserSchema" = Relationship(back_populates="api_keys")

    def to_model(self) -> None:
        pass



from fastapi import Request, HTTPException
from fastapi_nextauth_jwt import NextAuthJWTv4

from admyral.models.auth import AuthenticatedUser
from admyral.config.config import CONFIG, DISABLE_AUTH, AUTH_SECRET, ADMYRAL_ENV
from admyral.server.deps import get_admyral_store
from admyral.logger import get_logger


logger = get_logger(__name__)


JWT = NextAuthJWTv4(
    secret=AUTH_SECRET, csrf_prevention_enabled=False if ADMYRAL_ENV == "dev" else True
)


def validate_and_decrypt_jwt(request: Request) -> dict:
    return JWT(request)


async def authenticate(request: Request) -> AuthenticatedUser:
    if DISABLE_AUTH:
        return AuthenticatedUser(
            user_id=CONFIG.default_user_id,
            email=CONFIG.default_user_email,
        )

    # extract user id from authentication method
    if "x-api-key" in request.headers:
        user_id = await get_admyral_store().search_api_key_owner(
            request.headers["x-api-key"]
        )
        if not user_id:
            raise HTTPException(status_code=401, detail="Invalid API Key")
    else:
        try:
            decrypted_token = validate_and_decrypt_jwt(request)
        except Exception as e:
            logger.error(f"Failed to validate token: {e}")
            raise HTTPException(status_code=401, detail="Invalid or missing token")
        user_id = decrypted_token.get("sub") or decrypted_token.get("id")

    if not user_id:
        # Missing user id
        raise HTTPException(status_code=401, detail="Invalid token")

    # check user for existance in the database
    user = await get_admyral_store().get_user(user_id)
    if not user:
        # User not found
        raise HTTPException(status_code=401, detail="Invalid token")

    return AuthenticatedUser(user_id=user_id, email=user.email)
